#!/usr/bin/env python3

import os, sys, fnmatch
import xml.etree.ElementTree as ET

"""
A static protocol code generator for the task of creating the wire representation
of a list of events/requests
"""

wltype_to_ctypes = {
    "uint": "uint32_t ",
    "fixed": "uint32_t ",
    "int": "int32_t ",
    "object": "struct wp_objid ",
    "new_id": "struct wp_objid ",
    "string": "const char *",
    "fd": "int ",
}


def write_enum(ostream, iface_name, enum):
    enum_name = enum.attrib["name"]
    is_bitfield = "bitfield" in enum.attrib and enum.attrib["bitfield"] == "true"

    for entry in enum:
        if entry.tag != "entry":
            continue
        entry_name = entry.attrib["name"]
        entry_value = entry.attrib["value"]

        full_name = (iface_name + "_" + enum_name + "_" + entry_name).upper()
        print("#define {} {}".format(full_name, entry_value), file=ostream)


def is_exportable(func_name, export_list):
    for e in export_list:
        if fnmatch.fnmatchcase(func_name, e):
            return True
    return False


def write_func(ostream, iface_name, func, is_request, func_no, export_list):
    func_name = (
        iface_name + "_" + ("req" if is_request else "evt") + "_" + func.attrib["name"]
    )

    for_export = is_exportable(func_name, export_list)
    if not for_export:
        return

    c_sig = ["struct transfer_states *ts", "struct wp_objid " + iface_name + "_id"]
    w_args = []

    num_fd_args = 0
    num_reg_args = 0
    num_obj_args = 0
    num_new_args = 0
    num_stretch_args = 0
    for arg in func:
        if arg.tag != "arg":
            continue

        arg_name = arg.attrib["name"]
        arg_type = arg.attrib["type"]
        arg_interface = arg.attrib["interface"] if "interface" in arg.attrib else None
        if arg_type == "new_id" and arg_interface is None:
            # Special case, for wl_registry_bind
            c_sig.append("const char *interface")
            c_sig.append("uint32_t version")
            c_sig.append("struct wp_objid id")
            w_args.append(("interface", "string", None))
            w_args.append(("version", "uint", None))
            w_args.append((arg_name, "new_id", None))
            num_obj_args += 1
            num_new_args += 1
            num_reg_args += 3
            num_stretch_args += 1
            continue

        if arg_type == "array":
            c_sig.append("int " + arg_name + "_count")
            c_sig.append("const uint8_t *" + arg_name + "_val")
        else:
            c_sig.append(wltype_to_ctypes[arg_type] + arg_name)
        w_args.append((arg_name, arg_type, arg_interface))
        if arg_type == "fd":
            num_fd_args += 1
        else:
            num_reg_args += 1
        if arg_type == "object" or arg_type == "new_id":
            num_obj_args += 1
        if arg_type == "new_id":
            num_new_args += 1
        if arg_type in ("array", "string"):
            num_stretch_args += 1

    send_signature = "static void send_{}({}) ".format(func_name, ", ".join(c_sig))

    W = lambda *x: print(*x, file=ostream)

    # Write function definition
    W(send_signature + " {")
    W("\tts->fd_size = 0;")
    W("\tts->msg_space[0] = {}.id;".format(iface_name + "_id"))
    W("\tts->msg_size = 2;")

    tmp_names = ["ctx"]
    for i, (arg_name, arg_type, arg_interface) in enumerate(w_args):
        if arg_type == "array":
            raise NotImplementedError()
            continue
        elif arg_type == "fd":
            W("\tts->fd_space[ts->fd_size++] = {};".format(arg_name))
            continue
        elif arg_type == "string":
            W("\tserialize_string(ts, {});".format(arg_name))
            continue
        elif arg_type == "object" or arg_type == "new_id":
            W("\tts->msg_space[ts->msg_size++] = {}.id;".format(arg_name))
        elif arg_type == "int":
            W("\tts->msg_space[ts->msg_size++] = (uint32_t){};".format(arg_name))
        elif arg_type == "uint" or arg_type == "fixed":
            W("\tts->msg_space[ts->msg_size++] = {};".format(arg_name))
        else:
            raise KeyError(arg_type)

    W("\tts->msg_space[1] = ((uint32_t)ts->msg_size << 18) | {};".format(func_no))
    if is_request:
        W("\tts->send(ts, ts->app, ts->comp);")
    else:
        W("\tts->send(ts, ts->comp, ts->app);")

    W("}")


if __name__ == "__main__":
    req_file, dest = sys.argv[1:3]
    sources = sys.argv[3:]
    assert dest.endswith(".h")
    dest_shortname = dest[:-2]
    header_flag = dest_shortname.upper().replace("/", "_") + "_H"

    export_list = open(req_file).read().split("\n")

    with open(dest, "w") as ostream:
        W = lambda *x: print(*x, file=ostream)

        W("#ifndef {}".format(header_flag))
        W("#include <stddef.h>")
        W("#include <stdint.h>")
        W("#include <string.h>")
        W("struct test_state;")
        W("struct wp_objid { uint32_t id; };")
        W("struct transfer_states {")
        W("\tuint32_t msg_space[256];")
        W("\tint fd_space[16];")
        W("\tunsigned int msg_size;")
        W("\tunsigned int fd_size;")
        W("\tstruct test_state *app;")
        W("\tstruct test_state *comp;")
        W(
            "\tvoid (*send)(struct transfer_states *, struct test_state *src, struct test_state *dst);"
        )
        W("};")
        # note: this script assumes that serialize_string will be used
        W("static void serialize_string(struct transfer_states *ts, const char *str) {")
        W("\tif (str) {")
        W("\t\tsize_t slen = strlen(str) + 1;")
        W("\t\tts->msg_space[ts->msg_size] = (uint32_t)slen;")
        W("\t\tmemcpy(&ts->msg_space[ts->msg_size + 1], str, slen);")
        W("\t\tts->msg_size += ((uint32_t)slen + 0x7) >> 2;")
        W("\t} else {")
        W("\t\tts->msg_space[ts->msg_size++] = 0;")
        W("\t}")
        W("}")

        for source in sorted(sources):
            tree = ET.parse(source)
            root = tree.getroot()
            for interface in root:
                if interface.tag != "interface":
                    continue
                iface_name = interface.attrib["name"]

                func_data = []
                nreq, nevt = 0, 0
                for item in interface:
                    if item.tag == "enum":
                        write_enum(ostream, iface_name, item)
                    elif item.tag == "request":
                        write_func(ostream, iface_name, item, True, nreq, export_list)
                        nreq += 1
                    elif item.tag == "event":
                        write_func(ostream, iface_name, item, False, nevt, export_list)
                        nevt += 1
                    elif item.tag == "description":
                        pass
                    else:
                        raise Exception(item.tag)

        W("#endif /* {} */".format(header_flag))



#!/usr/bin/env python3

import sys, os, subprocess, shutil

"""
Script to check that Waypipe builds and that tests pass in all of
its configurations.
"""

waypipe_root, build_root = sys.argv[1], sys.argv[2]
os.makedirs(build_root, exist_ok=True)

setups = [
    ("regular", ["--buildtype", "debugoptimized"], {}),
    ("release", ["--buildtype", "release"], {}),
    ("clang", ["--buildtype", "debugoptimized"], {"CC": "clang"}),
    (
        "clang-tsan",
        ["--buildtype", "debugoptimized", "-Db_sanitize=thread"],
        {"CC": "clang"},
    ),
    (
        "clang-asan",
        ["--buildtype", "debugoptimized", "-Db_sanitize=address,undefined"],
        {"CC": "clang"},
    ),
    (
        "empty",
        [
            "--buildtype",
            "debugoptimized",
            "-Dwith_video=disabled",
            "-Dwith_lz4=disabled",
            "-Dwith_zstd=disabled",
            "-Dwith_dmabuf=disabled",
        ],
        {"CC": "gcc"},
    ),
    (
        "novideo",
        [
            "--buildtype",
            "debugoptimized",
            "-Dwith_video=disabled",
        ],
        {"CC": "gcc"},
    ),
    (
        "nolz4",
        [
            "--buildtype",
            "debugoptimized",
            "-Dwith_lz4=disabled",
        ],
        {"CC": "gcc"},
    ),
    (
        "unity",
        ["--buildtype", "debugoptimized", "--unity", "on", "--unity-size", "400"],
        {"CC": "gcc", "CFLAGS": "-pedantic -D_GNU_SOURCE"},
    ),
    (
        "error",
        ["--buildtype", "debugoptimized"],
        {"CC": "gcc", "CFLAGS": "-Wunused-result -std=c11 -pedantic -ggdb3 -O1"},
    ),
]
main_options = ["video", "dmabuf", "lz4", "zstd", "vaapi"]
bool_map = {True: "enabled", False: "disabled"}
for compiler in ["gcc", "clang"]:
    for flags in range(2 ** len(main_options)):
        bool_options = [(2**i) & flags != 0 for i in range(len(main_options))]
        name = "-".join(
            ["poly", compiler] + [m for m, b in zip(main_options, bool_options) if b]
        )
        flag_values = [
            "-Dwith_{}={}".format(m, bool_map[b])
            for m, b in zip(main_options, bool_options)
        ]
        setups.append(
            (name, ["--buildtype", "debugoptimized"] + flag_values, {"CC": compiler})
        )

if len(sys.argv) >= 4:
    setups = [(s, c, e) for s, c, e in setups if s == sys.argv[3]]

base_env = os.environ.copy()
for key, options, env in setups:
    print(key, end=" ")
    sys.stdout.flush()
    nenv = base_env.copy()
    for e in env:
        nenv[e] = env[e]

    bdir = os.path.join(build_root, key)
    try:
        shutil.rmtree(bdir)
    except FileNotFoundError:
        pass
    r1 = subprocess.run(
        ["meson", waypipe_root, bdir] + options, capture_output=True, env=nenv
    )
    if r1.returncode:
        print("failed")
        print(r1.stdout, r1.stderr, r1.returncode)
        continue
    r2 = subprocess.run(["ninja", "test"], cwd=bdir, capture_output=True, env=nenv)
    if r2.returncode:
        print("failed")
        print(r2.stdout, r2.stderr, r2.returncode)
        continue
    print("passed")



#!/usr/bin/env python3

import os, sys, fnmatch
import xml.etree.ElementTree as ET
import argparse

"""
A static protocol code generator.
"""

wltype_to_ctypes = {
    "uint": "uint32_t ",
    "fixed": "uint32_t ",
    "int": "int32_t ",
    "object": "struct wp_object *",
    "new_id": "struct wp_object *",
    "string": "const char *",
    "fd": "int ",
}


def superstring(a, b):
    na, nb = len(a), len(b)
    if nb > na:
        b, a, nb, na = a, b, na, nb
    # A contains B
    for i in range(na - nb + 1):
        if a[i : nb + i] == b:
            return a

    # suffix of B is prefix of A
    ba_overlap = 0
    for i in range(1, nb):
        if b[-i:] == a[:i]:
            ba_overlap = i

    # suffix of A is prefix of B
    ab_overlap = 0
    for i in range(1, nb):
        if a[-i:] == b[:i]:
            ab_overlap = i

    if ba_overlap > ab_overlap:
        return b + a[ba_overlap:]
    else:
        return a + b[ab_overlap:]


def get_offset(haystack, needle):
    for i in range(len(haystack) - len(needle) + 1):
        if haystack[i : i + len(needle)] == needle:
            return i
    return None


def shortest_superstring(strings):
    """
    Given strings L_1,...L_n over domain U, report an approximation
    of the shortest superstring of the lists, and offsets of the
    L_i into this string. Has O(n^3) runtime; O(n^2 polylog) is possible.
    """
    if not len(strings):
        return None

    pool = []
    for s in strings:
        if s not in pool:
            pool.append(s)

    while len(pool) > 1:
        max_overlap = 0
        best = None
        for i in range(len(pool)):
            for j in range(i):
                d = len(pool[i]) + len(pool[j]) - len(superstring(pool[i], pool[j]))
                if d >= max_overlap:
                    max_overlap = d
                    best = (j, i)

        s = superstring(pool[best[0]], pool[best[1]])
        del pool[best[1]]
        del pool[best[0]]
        pool.append(s)

    sstring = pool[0]
    for s in strings:
        assert get_offset(sstring, s) != None, ("substring property", sstring, s)

    return sstring


def write_enum(is_header, ostream, iface_name, enum):
    if not is_header:
        return

    enum_name = enum.attrib["name"]
    is_bitfield = "bitfield" in enum.attrib and enum.attrib["bitfield"] == "true"

    long_name = iface_name + "_" + enum_name
    print("enum " + long_name + " {", file=ostream)
    for entry in enum:
        if entry.tag != "entry":
            continue
        entry_name = entry.attrib["name"]
        entry_value = entry.attrib["value"]

        full_name = long_name.upper() + "_" + entry_name.upper()
        print("\t" + full_name + " = " + entry_value + ",", file=ostream)
    print("};", file=ostream)


def write_version(is_header, ostream, iface_name, version):
    if not is_header:
        return

    print(
        "#define " + iface_name.upper() + "_INTERFACE_VERSION " + str(version),
        file=ostream,
    )


def is_exportable(func_name, export_list):
    for e in export_list:
        if fnmatch.fnmatchcase(func_name, e):
            return True
    return False


def write_func(is_header, ostream, func_name, func):
    c_sig = ["struct context *ctx"]
    w_args = []

    num_fd_args = 0
    num_reg_args = 0
    num_obj_args = 0
    num_new_args = 0
    num_stretch_args = 0
    for arg in func:
        if arg.tag != "arg":
            continue

        arg_name = arg.attrib["name"]
        arg_type = arg.attrib["type"]
        arg_interface = arg.attrib["interface"] if "interface" in arg.attrib else None
        if arg_type == "new_id" and arg_interface is None:
            # Special case, for wl_registry_bind
            c_sig.append("const char *interface")
            c_sig.append("uint32_t version")
            c_sig.append("struct wp_object *id")
            w_args.append(("interface", "string", None))
            w_args.append(("version", "uint", None))
            w_args.append((arg_name, "new_id", None))
            num_obj_args += 1
            num_new_args += 1
            num_reg_args += 3
            num_stretch_args += 1
            continue

        if arg_type == "array":
            c_sig.append("uint32_t " + arg_name + "_count")
            c_sig.append("const uint8_t *" + arg_name + "_val")
        else:
            c_sig.append(wltype_to_ctypes[arg_type] + arg_name)
        w_args.append((arg_name, arg_type, arg_interface))
        if arg_type == "fd":
            num_fd_args += 1
        else:
            num_reg_args += 1
        if arg_type == "object" or arg_type == "new_id":
            num_obj_args += 1
        if arg_type == "new_id":
            num_new_args += 1
        if arg_type in ("array", "string"):
            num_stretch_args += 1

    do_signature = "void do_{}({});".format(func_name, ", ".join(c_sig))
    handle_signature = "static void call_{}(struct context *ctx, const uint32_t *payload, const int *fds, struct message_tracker *mt)".format(
        func_name
    )

    W = lambda *x: print(*x, file=ostream)
    if is_header:
        W(do_signature)
    if not is_header:
        # Write function definition
        W(do_signature)
        W(handle_signature + " {")
        if num_reg_args > 0:
            W("\tunsigned int i = 0;")
        if num_fd_args > 0:
            W("\tunsigned int k = 0;")

        tmp_names = ["ctx"]
        n_fds_left = num_fd_args
        n_reg_left = num_reg_args
        for i, (arg_name, arg_type, arg_interface) in enumerate(w_args):
            if arg_type == "array":
                n_reg_left -= 1
                W(
                    "\tconst uint8_t *arg{}_b = (const uint8_t *)&payload[i + 1];".format(
                        i
                    )
                )
                W("\tuint32_t arg{}_a = payload[i];".format(i))
                if n_reg_left > 0:
                    W("\ti += 1 + (unsigned int)((arg{}_a + 0x3) >> 2);".format(i))

                tmp_names.append("arg{}_a".format(i))
                tmp_names.append("arg{}_b".format(i))
                continue

            tmp_names.append("arg{}".format(i))

            if arg_type == "fd":
                n_fds_left -= 1
                W("\tint arg{} = fds[{}];".format(i, "k++" if n_fds_left > 0 else "k"))
                continue

            n_reg_left -= 1
            if arg_type == "string":
                W("\tconst char *arg{} = (const char *)&payload[i + 1];".format(i))
                W("\tif (!payload[i]) arg{} = NULL;".format(i))
                if n_reg_left > 0:
                    W("\ti += 1 + ((payload[i] + 0x3) >> 2);")
                continue

            i_incr = "i++" if n_reg_left > 0 else "i"

            if arg_type == "object" or arg_type == "new_id":
                if arg_interface is None:
                    intf_str = "NULL"
                else:
                    intf_str = "&intf_" + arg_interface
                W(
                    "\tstruct wp_object *arg{} = get_object(mt, payload[{}], {});".format(
                        i, i_incr, intf_str
                    )
                )
            elif arg_type == "int":
                W("\tint32_t arg{} = (int32_t)payload[{}];".format(i, i_incr))
            elif arg_type == "uint" or arg_type == "fixed":
                W("\tuint32_t arg{} = payload[{}];".format(i, i_incr))

        W("\tdo_{}({});".format(func_name, ", ".join(tmp_names)))
        if num_obj_args == 0:
            W("\t(void)mt;")
        if num_fd_args == 0:
            W("\t(void)fds;")
        if num_reg_args == 0:
            W("\t(void)payload;")

        W("}")


def load_msg_data(func_name, func, for_export):
    w_args = []
    for arg in func:
        if arg.tag != "arg":
            continue
        arg_name = arg.attrib["name"]
        arg_type = arg.attrib["type"]
        arg_interface = arg.attrib["interface"] if "interface" in arg.attrib else None
        if arg_type == "new_id" and arg_interface is None:
            w_args.append(("interface", "string", None))
            w_args.append(("version", "uint", None))
            w_args.append((arg_name, "new_id", None))
        else:
            w_args.append((arg_name, arg_type, arg_interface))

    new_objs = []
    for arg_name, arg_type, arg_interface in w_args:
        if arg_type == "new_id":
            new_objs.append(
                "&intf_" + arg_interface if arg_interface is not None else "NULL"
            )

    # gap coding: 0=end,1=new_obj,2=array,3=string
    num_fd_args = 0
    gaps = [0]
    gap_ends = []
    for arg_name, arg_type, arg_interface in w_args:
        if arg_type == "fd":
            num_fd_args += 1
            continue

        gaps[-1] += 1
        if arg_type in ("new_id", "string", "array"):
            gap_ends.append({"new_id": 1, "string": 3, "array": 2}[arg_type])
            gaps.append(0)
    gap_ends.append(0)
    gap_codes = [str(g * 4 + e) for g, e in zip(gaps, gap_ends)]

    is_destructor = "type" in func.attrib and func.attrib["type"] == "destructor"
    is_request = item.tag == "request"
    short_name = func.attrib["name"]

    return (
        is_request,
        func_name,
        short_name,
        new_objs,
        gap_codes,
        is_destructor,
        num_fd_args,
        for_export,
    )


def write_interface(
    ostream, iface_name, func_data, gap_code_array, new_obj_array, dest_name
):
    reqs, evts = [], []
    for x in func_data:
        if x[0]:
            reqs.append(x)
        else:
            evts.append(x)

    W = lambda *x: print(*x, file=ostream)

    if len(reqs) > 0 or len(evts) > 0:
        W("static const struct msg_data msgs_" + iface_name + "[] = {")

    msg_names = []
    for x in reqs + evts:
        (
            is_request,
            func_name,
            short_name,
            new_objs,
            gap_codes,
            is_destructor,
            num_fd_args,
            for_export,
        ) = x
        msg_names.append(short_name)

        mda = []
        mda.append(
            "gaps_{} + {}".format(dest_name, get_offset(gap_code_array, gap_codes))
        )
        if len(new_objs) > 0:
            mda.append(
                "objt_{} + {}".format(dest_name, get_offset(new_obj_array, new_objs))
            )
        else:
            mda.append("NULL")

        mda.append(("call_" + func_name) if for_export else "NULL")
        mda.append(str(num_fd_args))
        mda.append("true" if is_destructor else "false")

        W("\t{" + ", ".join(mda) + "},")

    mcn = "NULL"
    if len(reqs) > 0 or len(evts) > 0:
        W("};")
        mcn = "msgs_" + iface_name

    W("const struct wp_interface intf_" + iface_name + " = {")
    W("\t" + mcn + ",")
    W("\t" + str(len(reqs)) + ",")
    W("\t" + str(len(evts)) + ",")
    W('\t"{}",'.format(iface_name))
    W('\t"{}",'.format("\\0".join(msg_names)))
    W("};")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("mode", help="Either 'header' or 'data'.")
    parser.add_argument(
        "export_list", help="List of events/requests which need parsing."
    )
    parser.add_argument("output_file", help="C file to create.")
    parser.add_argument("protocols", nargs="+", help="XML protocol files to use.")
    args = parser.parse_args()

    is_header = {"data": False, "header": True}[args.mode]
    if is_header:
        assert args.output_file[-2:] == ".h"
    else:
        assert args.output_file[-2:] == ".c"
    dest_name = os.path.basename(args.output_file)[:-2].replace("-", "_")

    export_list = open(args.export_list).read().split("\n")

    intfset = set()
    for source in args.protocols:
        tree = ET.parse(source)
        root = tree.getroot()
        for intf in root:
            if intf.tag == "interface":
                intfset.add(intf.attrib["name"])
                for msg in intf:
                    for arg in msg:
                        if "interface" in arg.attrib:
                            intfset.add(arg.attrib["interface"])
    interfaces = sorted(intfset)

    header_guard = "{}_H".format(dest_name.upper())
    with open(args.output_file, "w") as ostream:
        W = lambda *x: print(*x, file=ostream)

        if is_header:
            W("#ifndef {}".format(header_guard))
            W("#define {}".format(header_guard))
            W()
        W('#include "symgen_types.h"')
        if not is_header:
            W("#include <stddef.h>")

        for intf in interfaces:
            W("extern const struct wp_interface intf_{};".format(intf))

        gap_code_list = []
        new_obj_list = []

        interface_data = []

        for source in sorted(args.protocols):
            tree = ET.parse(source)
            root = tree.getroot()
            for interface in root:
                if interface.tag != "interface":
                    continue
                iface_name = interface.attrib["name"]

                write_version(
                    is_header, ostream, iface_name, interface.attrib["version"]
                )

                func_data = []
                for item in interface:
                    if item.tag == "enum":
                        write_enum(is_header, ostream, iface_name, item)
                    elif item.tag == "request" or item.tag == "event":
                        is_req = item.tag == "request"
                        func_name = (
                            iface_name
                            + "_"
                            + ("req" if is_req else "evt")
                            + "_"
                            + item.attrib["name"]
                        )

                        for_export = is_exportable(func_name, export_list)
                        if for_export:
                            write_func(is_header, ostream, func_name, item)
                        if not is_header:
                            func_data.append(load_msg_data(func_name, item, for_export))

                    elif item.tag == "description":
                        pass
                    else:
                        raise Exception(item.tag)

                for x in func_data:
                    gap_code_list.append(x[4])
                    new_obj_list.append(x[3])

                interface_data.append((iface_name, func_data))

        if not is_header:
            gap_code_array = shortest_superstring(gap_code_list)
            new_obj_array = shortest_superstring(new_obj_list)

            if new_obj_array is not None:
                W("static const struct wp_interface *objt_" + dest_name + "[] = {")
                W("\t" + ",\n\t".join(new_obj_array))
                W("};")

            if gap_code_array is not None:
                W("static const uint16_t gaps_" + dest_name + "[] = {")
                W("\t" + ",\n\t".join(gap_code_array))
                W("};")

            for iface_name, func_data in interface_data:
                write_interface(
                    ostream,
                    iface_name,
                    func_data,
                    gap_code_array,
                    new_obj_array,
                    dest_name,
                )

        if is_header:
            W()
            W("#endif /* {} */".format(header_guard))



#!/usr/bin/env python3

"""
Verifying all the ways in which waypipe can fail before even making a connection.
"""

if __name__ != "__main__":
    quit(1)

import os, subprocess, time, signal, socket


def try_unlink(path):
    try:
        os.unlink(path)
    except FileNotFoundError:
        pass


def make_socket(path):
    folder, filename = os.path.split(path)
    cwdir = os.open(".", os.O_RDONLY | os.O_DIRECTORY)

    display_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    os.chdir(folder)
    display_socket.bind(filename)
    display_socket.listen()
    os.fchdir(cwdir)
    os.close(cwdir)

    return display_socket


waypipe_path = os.environ["TEST_WAYPIPE_PATH"]
sleep_path = os.environ["TEST_SLEEP_PATH"]
fake_ssh_path = os.environ["TEST_FAKE_SSH_PATH"]
ld_library_path = (
    os.environ["LD_LIBRARY_PATH"] if "LD_LIBRARY_PATH" in os.environ else ""
)

xdg_runtime_dir = os.path.abspath("./run/")
os.makedirs(xdg_runtime_dir, mode=0o700, exist_ok=True)
os.chmod(xdg_runtime_dir, 0o700)

all_succeeding = True

wayland_display = "wayland-display"
client_socket_path = xdg_runtime_dir + "/client-socket"
server_socket_path = xdg_runtime_dir + "/server-socket"
ssh_socket_path = xdg_runtime_dir + "/ssh-socket"
wayland_display_path = xdg_runtime_dir + "/" + wayland_display

try_unlink(wayland_display_path)
display_socket = make_socket(wayland_display_path)

USE_SOCKETPAIR = 1 << 1
EXPECT_SUCCESS = 1 << 2
EXPECT_TIMEOUT = 1 << 3
EXPECT_FAILURE = 1 << 4


def run_test(name, command, env, flags):
    try_unlink(client_socket_path)
    try_unlink(server_socket_path)
    try_unlink(server_socket_path + ".disp.sock")
    if flags & USE_SOCKETPAIR:
        sockets = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM)
        conn_socket = 999
        os.dup2(sockets[1].fileno(), conn_socket, inheritable=True)
        env = dict(env, WAYLAND_SOCKET=str(conn_socket))
        pfds = [conn_socket]
    else:
        pfds = []

    timed_out = False
    log_path = os.path.join(xdg_runtime_dir, "sfail_{}.txt".format(name))
    logfile = open(log_path, "wb")
    print(env, " ".join(command))

    proc = subprocess.Popen(
        command,
        env=env,
        stdin=subprocess.DEVNULL,
        stdout=logfile,
        stderr=subprocess.STDOUT,
        pass_fds=pfds,
        start_new_session=True,
    )
    try:
        output, none = proc.communicate(timeout=1.0)
    except subprocess.TimeoutExpired as e:
        # Program is waiting indefinitely for something.
        # Kill it, and all children.
        pgrp = os.getpgid(proc.pid)
        os.killpg(pgrp, signal.SIGKILL)
        retcode = None
        timed_out = True
    else:
        retcode = proc.returncode

    logfile.close()
    output = open(log_path, "rb").read()

    if flags & USE_SOCKETPAIR:
        os.close(conn_socket)

    log_path = os.path.join(xdg_runtime_dir, "weston_out.txt")
    with open(log_path, "wb") as out:
        out.write(output)

    result = (
        "timeout"
        if timed_out
        else ("fail({})".format(retcode) if retcode != 0 else "pass")
    )

    global all_succeeding
    if flags & EXPECT_SUCCESS:
        if timed_out or retcode != 0:
            print(
                "Run {} failed when it should have succeeded".format(name),
                output,
                retcode,
                "timeout" if timed_out else "notimeout",
            )
            all_succeeding = False
        else:
            print("Run {} passed.".format(name), output)
    elif flags & EXPECT_FAILURE:
        if timed_out or retcode == 0:
            print(
                "Run {} succeeded when it should have failed".format(name),
                output,
                retcode,
                "timeout" if timed_out else "notimeout",
            )
            all_succeeding = False
        else:
            print("Run {} passed.".format(name), output)
    elif flags & EXPECT_TIMEOUT:
        if not timed_out:
            print(
                "Run {} stopped when it should have continued".format(name),
                output,
                retcode,
            )
            all_succeeding = False
        else:
            print("Run {} passed.".format(name), output)
    else:
        raise NotImplementedError


wait_cmd = [sleep_path, "10.0"]
invalid_hostname = "@"
fake_ssh_dir = os.path.dirname(fake_ssh_path)
waypipe_dir = os.path.dirname(waypipe_path)

base_env = {"LD_LIBRARY_PATH": ld_library_path, "PATH": ""}
standard_env = dict(base_env, XDG_RUNTIME_DIR=xdg_runtime_dir)
ssh_only_env = dict(standard_env, PATH=fake_ssh_dir)
ssh_env = dict(standard_env, PATH=fake_ssh_dir + ":" + waypipe_dir)
# Configurations that should fail
run_test(
    "b_client_long_disp",
    [waypipe_path, "-s", client_socket_path, "client"],
    dict(base_env, WAYLAND_DISPLAY=("/" + "x" * 107)),
    EXPECT_FAILURE,
)
run_test(
    "b_client_disp_dne",
    [waypipe_path, "-s", client_socket_path, "client"],
    dict(base_env, WAYLAND_DISPLAY=xdg_runtime_dir + "/dne"),
    EXPECT_FAILURE,
)
run_test(
    "b_client_no_env",
    [waypipe_path, "-s", client_socket_path, "client"],
    base_env,
    EXPECT_FAILURE,
)
run_test(
    "b_server_oneshot_no_env",
    [waypipe_path, "-o", "-s", server_socket_path, "server"] + wait_cmd,
    base_env,
    EXPECT_TIMEOUT,
)
run_test(
    "b_client_bad_pipe1",
    [waypipe_path, "-s", client_socket_path, "client"],
    dict(base_env, WAYLAND_SOCKET="33"),
    EXPECT_FAILURE,
)
run_test(
    "b_client_bad_pipe2",
    [waypipe_path, "-s", client_socket_path, "client"],
    dict(base_env, WAYLAND_SOCKET="777777777777777777777777777"),
    EXPECT_FAILURE,
)
run_test(
    "b_client_bad_pipe3",
    [waypipe_path, "-s", client_socket_path, "client"],
    dict(base_env, WAYLAND_SOCKET="0x33"),
    EXPECT_FAILURE,
)
run_test(
    "b_client_nxdg_offset",
    [waypipe_path, "-s", client_socket_path, "client"],
    dict(base_env, WAYLAND_DISPLAY=wayland_display),
    EXPECT_FAILURE,
)
run_test(
    "b_server_no_env",
    [waypipe_path, "-s", server_socket_path, "server"] + wait_cmd,
    base_env,
    EXPECT_FAILURE,
)
run_test(
    "g_ssh_test_nossh_env",
    [waypipe_path, "-o", "-s", ssh_socket_path, "ssh", invalid_hostname] + wait_cmd,
    dict(standard_env, WAYLAND_DISPLAY=wayland_display),
    EXPECT_FAILURE,
)


# Configurations that should succeed
run_test(
    "g_help",
    [waypipe_path, "--help"],
    base_env,
    EXPECT_SUCCESS,
)
run_test(
    "g_server_std_env",
    [waypipe_path, "-s", server_socket_path, "server"] + wait_cmd,
    standard_env,
    EXPECT_TIMEOUT,
)
run_test(
    "g_client_std_env",
    [waypipe_path, "-s", client_socket_path, "client"],
    dict(standard_env, WAYLAND_DISPLAY=wayland_display_path),
    EXPECT_TIMEOUT,
)
run_test(
    "g_client_offset_sock",
    [waypipe_path, "-s", client_socket_path, "client"],
    dict(standard_env, WAYLAND_DISPLAY=wayland_display),
    EXPECT_TIMEOUT,
)
run_test(
    "g_client_pipe_env",
    [waypipe_path, "-s", client_socket_path, "client"],
    dict(standard_env),
    EXPECT_TIMEOUT | USE_SOCKETPAIR,
)
run_test(
    "g_ssh_test_oneshot",
    [waypipe_path, "-o", "-s", ssh_socket_path, "ssh", invalid_hostname] + wait_cmd,
    dict(ssh_env, WAYLAND_DISPLAY=wayland_display),
    EXPECT_TIMEOUT,
)
run_test(
    "g_ssh_test_reg",
    [waypipe_path, "-s", ssh_socket_path, "ssh", invalid_hostname] + wait_cmd,
    dict(ssh_env, WAYLAND_DISPLAY=wayland_display),
    EXPECT_TIMEOUT,
)
run_test(
    "g_ssh_test_remotebin",
    [
        waypipe_path,
        "--oneshot",
        "--remote-bin",
        waypipe_path,
        "-s",
        ssh_socket_path,
        "ssh",
        invalid_hostname,
    ]
    + wait_cmd,
    dict(ssh_only_env, WAYLAND_DISPLAY=wayland_display),
    EXPECT_TIMEOUT,
)

try_unlink(client_socket_path)
try_unlink(wayland_display_path)
quit(0 if all_succeeding else 1)



#!/usr/bin/env python3

if __name__ != "__main__":
    quit(1)

import os, subprocess, time, signal
import multiprocessing


def try_unlink(path):
    try:
        os.unlink(path)
    except FileNotFoundError:
        pass


def wait_until_exists(path):
    for i in range(100):
        if os.path.exists(path):
            return True
        time.sleep(0.01)
    else:
        return False


def safe_cleanup(process):
    assert type(process) == subprocess.Popen
    for i in range(3):
        if process.poll() is None:
            # certain weston client programs appear to initiate shutdown proceedings correctly; however, they appear to wait for a frame beforehand, and the headless weston doesn't ask for additional frames
            process.send_signal(signal.SIGINT)
            time.sleep(0.5)
    try:
        process.wait(100)
    except subprocess.TimeoutExpired:
        process.kill()
        try:
            process.wait(1)
        except subprocess.TimeoutExpired:
            # no third chances
            process.terminate()


weston_path = os.environ["TEST_WESTON_PATH"]
waypipe_path = os.environ["TEST_WAYPIPE_PATH"]
ld_library_path = (
    os.environ["LD_LIBRARY_PATH"] if "LD_LIBRARY_PATH" in os.environ else ""
)

sub_tests = {
    "SHM": ["TEST_WESTON_SHM_PATH"],
    "EGL": ["TEST_WESTON_EGL_PATH", "-o"],
    "DMABUF": ["TEST_WESTON_DMA_PATH"],
    "TERM": ["TEST_WESTON_TERM_PATH"],
    "PRES": ["TEST_WESTON_PRES_PATH"],
    "SUBSURF": ["TEST_WESTON_SUBSURF_PATH"],
}
for k, v in list(sub_tests.items()):
    if v[0] in os.environ:
        v[0] = os.environ[v[0]]
    else:
        del sub_tests[k]

xdg_runtime_dir = os.path.abspath("./run/")


# weston does not currently appear to support setting absolute socket paths
socket_path = "w_sock"
abs_socket_path = os.path.join(xdg_runtime_dir, socket_path)

mainenv = {"XDG_RUNTIME_DIR": xdg_runtime_dir, "LD_LIBRARY_PATH": ld_library_path}

weston_command = [
    weston_path,
    "--backend=headless-backend.so",
    "--socket=" + socket_path,
    # "--use-pixman",
    "--width=1111",
    "--height=777",
]

arguments = subprocess.check_output([weston_path, "--help"]).decode()
if "--use-gl" in arguments:
    weston_command.append("--use-gl")

try:
    import psutil
except ImportError:
    psutil = None

nontrivial_failures = False

subenv = {
    "WAYLAND_DISPLAY": abs_socket_path,
    "WAYLAND_DEBUG": "1",
    "XDG_RUNTIME_DIR": xdg_runtime_dir,
    "LD_LIBRARY_PATH": ld_library_path,
    "ASAN_OPTIONS": "detect_leaks=0",
}

wp_serv_env = {
    "WAYLAND_DEBUG": "1",
    "XDG_RUNTIME_DIR": xdg_runtime_dir,
    "LD_LIBRARY_PATH": ld_library_path,
    "ASAN_OPTIONS": "detect_leaks=0",
}

subproc_args = {"env": subenv, "stdin": subprocess.DEVNULL, "stderr": subprocess.STDOUT}

wp_serv_args = {
    "env": wp_serv_env,
    "stdin": subprocess.DEVNULL,
    "stderr": subprocess.STDOUT,
}


def get_child_process(proc_pid, expected_name, sub_test_name):
    if psutil is not None:
        # assuming pid has not been recycled/duplicated
        proc = psutil.Process(proc_pid)
        if proc.name() == "waypipe":
            for i in range(5):
                kids = proc.children()
                if len(kids) > 0:
                    break
                time.sleep(0.01)
            else:
                print(
                    "For test",
                    sub_test_name,
                    "waypipe server's command may have crashed",
                )
            if len(kids) == 1:
                wp_child = kids[0]
                try:
                    if wp_child.name() != expected_name:
                        print(
                            "Unusual child process name",
                            wp_child.name(),
                            "does not match",
                            expected_name,
                        )
                except psutil.NoSuchProcess:
                    pass


def open_logfile(name):
    path = os.path.join(xdg_runtime_dir, name)
    return path, open(path, "wb")


def start_waypipe(socket_path, control_path, logfile, command, oneshot):
    prefix = [waypipe_path, "--debug", "--socket", socket_path]
    if oneshot:
        prefix += ["--oneshot"]
    client_command = prefix + ["client"]
    server_command = prefix + ["--control", control_path, "server"] + command
    client = subprocess.Popen(client_command, stdout=logfile, **subproc_args)
    if not wait_until_exists(socket_path):
        raise Exception("The waypipe socket file at " + socket_path + " did not appear")
    server = subprocess.Popen(server_command, stdout=logfile, **wp_serv_args)
    return server, client


def cleanup_oneshot(client, server, child):
    if child is not None:
        try:
            child.send_signal(signal.SIGINT)
        except psutil.NoSuchProcess:
            time.sleep(0.1)
            safe_cleanup(server)
            time.sleep(0.1)
            safe_cleanup(client)
        else:
            server.wait()
            client.wait()
    else:
        safe_cleanup(server)
        time.sleep(0.1)
        safe_cleanup(client)
    return client.returncode, server.returncode


def cleanup_multi(client, server, child):
    if child is not None:
        try:
            child.send_signal(signal.SIGINT)
        except psutil.NoSuchProcess:
            pass
    time.sleep(0.1)
    safe_cleanup(server)
    time.sleep(0.1)
    safe_cleanup(client)
    return client.returncode, server.returncode


def run_sub_test(args):
    sub_test_name, command = args
    nontrivial_failures = False

    ocontrol_path = os.path.join(xdg_runtime_dir, sub_test_name + "_octrl")
    mcontrol_path = os.path.join(xdg_runtime_dir, sub_test_name + "_mctrl")
    owp_socket_path = os.path.join(xdg_runtime_dir, sub_test_name + "_osocket")
    mwp_socket_path = os.path.join(xdg_runtime_dir, sub_test_name + "_msocket")
    try_unlink(owp_socket_path)
    try_unlink(mwp_socket_path)
    try_unlink(ocontrol_path)
    try_unlink(mcontrol_path)

    ref_log_path, ref_out = open_logfile(sub_test_name + "_ref_out.txt")
    ref_proc = subprocess.Popen(command, stdout=ref_out, **subproc_args)

    owp_log_path, owp_out = open_logfile(sub_test_name + "_owp_out.txt")
    mwp_log_path, mwp_out = open_logfile(sub_test_name + "_mwp_out.txt")

    owp_server, owp_client = start_waypipe(
        owp_socket_path, ocontrol_path, owp_out, command, True
    )
    mwp_server, mwp_client = start_waypipe(
        mwp_socket_path, mcontrol_path, mwp_out, command, False
    )

    owp_child = get_child_process(
        owp_server.pid, os.path.basename(command[0]), sub_test_name
    )
    mwp_child = get_child_process(
        mwp_server.pid, os.path.basename(command[0]), sub_test_name
    )

    print("Launched", sub_test_name)

    time.sleep(1)

    # Verify that replacing the control pipe (albeit with itself) doesn't break anything
    # (Since the connection is a unix domain socket, almost no packets will be in flight,
    # so the test isn't that comprehensive)
    print("Resetting", sub_test_name)
    open(ocontrol_path, "w").write(owp_socket_path)
    open(mcontrol_path, "w").write(mwp_socket_path)
    try_unlink(ocontrol_path)
    try_unlink(mcontrol_path)

    time.sleep(1)

    print("Closing", sub_test_name)

    # Beware sudden PID reuse...
    safe_cleanup(ref_proc)
    ref_out.close()

    occode, oscode = cleanup_oneshot(owp_client, owp_server, owp_child)
    mccode, mscode = cleanup_multi(mwp_client, mwp_server, mwp_child)

    try_unlink(owp_socket_path)
    try_unlink(mwp_socket_path)
    owp_out.close()
    mwp_out.close()

    # -2, because applications sometimes return with the sigint error
    if ref_proc.returncode not in (0, -2):
        print(
            "Test {}, run directly, failed (code={}). See logfile at {}".format(
                sub_test_name, ref_proc.returncode, ref_log_path
            )
        )
    else:
        if oscode in (0, -2) and occode == 0:
            print("Oneshot test", sub_test_name, "passed")
        else:
            print(
                "Oneshot test {}, run indirectly, failed (ccode={} scode={}). See logfile at {}".format(
                    sub_test_name, occode, oscode, owp_log_path
                )
            )
            nontrivial_failures = True
        if mscode in (0, -2) and mccode in (0, -2):
            print("Regular test", sub_test_name, "passed")
        else:
            print(
                "Regular test {}, run indirectly, failed (ccode={} scode={}). See logfile at {}".format(
                    sub_test_name, mccode, mscode, mwp_log_path
                )
            )
            nontrivial_failures = True
    return nontrivial_failures


os.makedirs(xdg_runtime_dir, mode=0o700, exist_ok=True)
os.chmod(xdg_runtime_dir, 0o700)
try_unlink(abs_socket_path)
try_unlink(abs_socket_path + ".lock")

weston_log_path = os.path.join(xdg_runtime_dir, "weston_out.txt")
weston_out = open(weston_log_path, "wb")
weston_proc = subprocess.Popen(
    weston_command,
    env=mainenv,
    stdin=subprocess.DEVNULL,
    stdout=weston_out,
    stderr=subprocess.STDOUT,
)


# Otherwise it's a race between weston and the clients
if not wait_until_exists(abs_socket_path):
    raise Exception(
        "weston failed to create expected display socket path, " + abs_socket_path
    )

with multiprocessing.Pool(3) as pool:
    nontriv_failures = pool.map(run_sub_test, [(k, v) for k, v in sub_tests.items()])

safe_cleanup(weston_proc)
weston_out.close()
if weston_proc.returncode != 0:
    print("Running headless weston failed. See logfile at ", weston_log_path)

if any(nontriv_failures):
    quit(1)
quit(0)
